/******************************************************************************
 *  CVS version:
 *     $Id: symbol.c,v 1.2 2005/04/07 11:22:47 nickie Exp $
 ******************************************************************************
 *
 *  C code file : symbol.c
 *  Project     : Llama Compiler
 *  Version     : 1.0 alpha
 *  Description : Generic symbol table
 *
 *  Comments: (in Greek iso-8859-7)
 *  ---------
 *    .
 *       .
 *      .
 *    
 */


/* ---------------------------------------------------------------------
   ---------------------------- Header files ---------------------------
   --------------------------------------------------------------------- */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "general.h"
#include "error.h"
#include "symbol.h"


/* ---------------------------------------------------------------------
   --------------------------   --------------------------
   --------------------------------------------------------------------- */

struct SymbolTable_tag {
   /*        */
   unsigned int  hashTableSize;      /*    */
   SymbolEntry * hashTable;          /*          */
   Scope         currentScope;       /*                */

   /*  ,   */
};

struct SymbolEntry_tag {
   /*        */
   Identifier   id;                 /*              */
   Scope        scope;              /*            */
   SymbolEntry  nextInHash;         /*    ..       */
   SymbolEntry  nextInScope;        /*       */

   /*  ,   */
};

struct Scope_tag {
   /*        */
   Scope        parent;                       /*   */
   SymbolEntry  entries;                      /*    */
   unsigned int nesting;                      /*       */
   bool         hidden;                       /*      */

   /*  ,   */
};


/* ---------------------------------------------------------------------
   ------        ------
   --------------------------------------------------------------------- */

SymbolTable symbol_make (unsigned int size)
{
   SymbolTable result = new(sizeof(struct SymbolTable_tag));
   unsigned int i;

   result->hashTableSize = size;
   result->hashTable = new(size * sizeof(SymbolEntry));
   result->currentScope = NULL;

   for (i = 0; i < size; i++)
      result->hashTable[i] = NULL;

   return result;
}

Scope scope_open (SymbolTable table)
{
   Scope result = new(sizeof(struct Scope_tag));

   ASSERT(table != NULL);

   result->parent = table->currentScope;
   result->entries = NULL;
   result->nesting =
      table->currentScope != NULL ? table->currentScope->nesting + 1 : 1;
   result->hidden = false;
   table->currentScope = result;
   return result;
}

Scope scope_close (SymbolTable table)
{
   Scope result = table->currentScope;
   SymbolEntry e;

   ASSERT(table != NULL);
   ASSERT(result != NULL);

   for (e = result->entries; e != NULL; e = e->nextInScope) {
      unsigned int hashValue = (unsigned int) e->id % table->hashTableSize;

      ASSERT(table->hashTable[hashValue] == e);
      table->hashTable[hashValue] = e->nextInHash;
      e->nextInHash = NULL;
   }

   table->currentScope = result->parent;
   return result;
}

void scope_hide (Scope scope, bool flag)
{
   ASSERT(scope != NULL);
   scope->hidden = flag;
}

void scope_insert (SymbolTable table, Scope scope)
{
   SymbolEntry e;
   SymbolEntry last = NULL;

   ASSERT(table != NULL);
   ASSERT(scope != NULL);

   for (e = scope->entries; e != NULL; e = e->nextInScope) {
      unsigned int hashValue = (unsigned int) e->id % table->hashTableSize;

      e->scope = table->currentScope;
      e->nextInHash = table->hashTable[hashValue];
      table->hashTable[hashValue] = e;
      last = e;
   }
   if (table->currentScope != NULL && last != NULL) {
      last->nextInScope = table->currentScope->entries;
      table->currentScope->entries = scope->entries;
   }
}

SymbolEntry symbol_enter (SymbolTable table, Identifier id, bool err)
{
   unsigned int hashValue = (unsigned int) id % table->hashTableSize;
   SymbolEntry e;

   ASSERT(table != NULL);

   /*        */

   if (err) {
      unsigned int currentNesting =
         table->currentScope != NULL ? table->currentScope->nesting : 0;

      for (e = table->hashTable[hashValue]; e != NULL; e = e->nextInHash) {
         unsigned int eNesting = e->scope != NULL ? e->scope->nesting : 0;

         if (eNesting < currentNesting)
            break;

         ASSERT(eNesting == currentNesting);
         if (e->id == id) {
            error("duplicate identifier: %s", id_name(id));
            return NULL;
         }
      }
   }

   /*  */

   e = new(sizeof(struct SymbolEntry_tag));
   e->id = id;
   e->scope = table->currentScope;
   e->nextInHash = table->hashTable[hashValue];
   table->hashTable[hashValue] = e;
   if (table->currentScope != NULL) {
      e->nextInScope = table->currentScope->entries;
      table->currentScope->entries = e;
   }
   else
      e->nextInScope = NULL;
   return e;
}

SymbolEntry symbol_lookup (SymbolTable table, Identifier id,
                           LookupType type, bool err)
{
   unsigned int currentNesting =
      table->currentScope != NULL ? table->currentScope->nesting : 0;
   unsigned int hashValue = (unsigned int) id % table->hashTableSize;
   SymbolEntry e;

   ASSERT(table != NULL);

   switch (type) {
      case LOOKUP_CURRENT_SCOPE:
         for (e = table->hashTable[hashValue]; e != NULL; e = e->nextInHash) {
            unsigned int eNesting = e->scope != NULL ? e->scope->nesting : 0;

            if (eNesting < currentNesting)
               break;

            ASSERT(eNesting == currentNesting);
            if (e->id == id)
               return e;
         }
         break;
      case LOOKUP_ALL_SCOPES:
         for (e = table->hashTable[hashValue]; e != NULL; e = e->nextInHash) {
            if (e->scope != NULL && e->scope->hidden)
               continue;
            if (e->id == id)
               return e;
         }
         break;
   }

   /* ,    */

   if (err)
      error("unknown identifier: %s", id_name(id));
   return NULL;
}
